@@ -1,3 +1,4 @@
+{-# OPTIONS_HADDOCK prune #-}
 {-|
 This module provides the high level functions that are re-exported
 by @Database.Postgres.Temp@. Additionally it includes some
@@ -8,9 +9,10 @@ module Database.Postgres.Temp.Internal where
 import Database.Postgres.Temp.Internal.Core
 import Database.Postgres.Temp.Internal.Config
 
+import qualified Control.Concurrent.Async as Async
 import           Control.DeepSeq
 import           Control.Exception
-import           Control.Monad (void)
+import           Control.Monad (void, join)
 import           Control.Monad.Trans.Cont
 import           Data.ByteString (ByteString)
 import qualified Data.Map.Strict as Map
@@ -20,6 +22,7 @@ import           System.Exit (ExitCode(..))
 import           System.IO.Unsafe (unsafePerformIO)
 import           System.Process
 import           Text.PrettyPrint.ANSI.Leijen hiding ((<$>))
+import           System.Directory
 
 -- | Handle for holding temporary resources, the @postgres@ process handle
 --   and @postgres@ connection information. The 'DB' also includes the
@@ -134,7 +137,7 @@ fastPostgresConfig =
 The default configuration. This will create a database called \"postgres\"
    via @initdb@ (it's default behavior).
    It will create a temporary directory for the data and a temporary directory
- for a unix socket on a random port.
+ for a unix socket and listen on 127.0.0.1 and ::1 on a random port.
  Additionally it will use the following \"postgresql.conf\"
  which is optimized for performance.
 
@@ -247,10 +250,13 @@ verbosePostgresConfig =
   , ("log_connections", "on")
   , ("log_disconnections", "on")
   , ("client_min_messages", "WARNING")
+  , ("log_min_messages", "WARNING")
+  , ("log_min_error_statement", "WARNING")
+  , ("log_statement", "all")
   ]
 
 {-|
-The similar to 'defaultConfig' log as much as possible.
+This is similar to 'defaultConfig' but it logs as much as possible..
 
 @since 1.21.0.0
 -}
@@ -259,7 +265,6 @@ verboseConfig = defaultConfig <> mempty
   { logger = pure print
   , postgresConfigFile = verbosePostgresConfig
   , initDbConfig = pure standardProcessConfig
-  , createDbConfig = pure standardProcessConfig
   , postgresConfig = standardProcessConfig
   }
 
@@ -326,9 +331,8 @@ start = startConfig defaultConfig
 --
 --   @since 1.12.0.0
 stop :: DB -> IO ()
-stop DB {..} = do
-  void $ stopPlan dbPostgresProcess
-  cleanupConfig dbResources
+stop DB {..} =
+  Async.concurrently_ (stopPlan dbPostgresProcess) $ cleanupConfig dbResources
 
 -- | Only stop the @postgres@ process but leave any temporary resources.
 --   Useful for testing backup strategies when used in conjunction with
@@ -465,6 +469,15 @@ cowCheck = unsafePerformIO $ do
        || missingFile ==  take (length missingFile) errorOutput
 {-# NOINLINE cowCheck #-}
 
+cpFlags :: String
+cpFlags = if cowCheck
+#ifdef darwin_HOST_OS
+  then "cp -Rc "
+#else
+  then "cp -R --reflink=auto "
+#endif
+  else "cp -R "
+
 {-|
 'defaultCacheConfig' attempts to determine if the @cp@ on the path
 supports \"copy on write\" flags and if it does, sets 'cacheUseCopyOnWrite'
@@ -571,27 +584,19 @@ newtype Snapshot = Snapshot { unSnapshot :: CompleteDirectoryType }
 {- |
 Shutdown the database and copy the directory to a folder.
 
-@since 1.20.0.0
+@since 1.29.0.0
 -}
 takeSnapshot
-  :: DirectoryType
-  -- ^ Either a 'Temporary' or preexisting 'Permanent' directory.
-  -> DB
+  :: DB
   -- ^ The handle. The @postgres@ is shutdown and the data directory is copied.
   -> IO (Either StartError Snapshot)
-takeSnapshot directoryType db = try $ do
+takeSnapshot db = try $ do
   throwIfNotSuccess id =<< stopPostgresGracefully db
-  let
-#ifdef darwin_HOST_OS
-    cpFlags = if cowCheck then "cp -Rc " else "cp -R "
-#else
-    cpFlags = if cowCheck then "cp -R --reflink=auto " else "cp -R "
-#endif
   bracketOnError
     (setupDirectoryType
       (toTemporaryDirectory db)
       "tmp-postgres-snapshot"
-      directoryType
+      Temporary
     )
     cleanupDirectoryType $ \snapShotDir -> do
       let snapshotCopyCmd = cpFlags <>
@@ -626,18 +631,28 @@ withDbCache $ \\cache -> withConfig (cacheConfig cache) $ \\db ->
     withConfig (snapshotConfig db) $ \\migratedDb -> ...
 @
 
-@since 1.20.0.0
+@since 1.29.0.0
 -}
 withSnapshot
-  :: DirectoryType
-  -> DB
+  :: DB
   -> (Snapshot -> IO a)
   -> IO (Either StartError a)
-withSnapshot dirType db f = bracket
-  (takeSnapshot dirType db)
+withSnapshot db f = bracket
+  (takeSnapshot db)
   (either mempty cleanupSnapshot)
   (either (pure . Left) (fmap Right . f))
 
+-- Helper for 'snapshotConfig' and 'cacheAction'
+fromFilePathConfig :: FilePath -> Config
+fromFilePathConfig filePath = mempty
+  { copyConfig = pure $ pure CopyDirectoryCommand
+      { sourceDirectory = filePath
+      , destinationDirectory = Nothing
+      , useCopyOnWrite = cowCheck
+      }
+  , initDbConfig = Zlich
+  }
+
 {-|
 Convert a snapshot into a 'Config' that includes a 'copyConfig' for copying the
 snapshot directory to a temporary directory.
@@ -645,11 +660,50 @@ snapshot directory to a temporary directory.
 @since 1.20.0.0
 -}
 snapshotConfig :: Snapshot -> Config
-snapshotConfig (Snapshot savePointPath) = mempty
-  { copyConfig = pure $ pure CopyDirectoryCommand
-      { sourceDirectory = toFilePath savePointPath
-      , destinationDirectory = Nothing
-      , useCopyOnWrite = cowCheck
-      }
-  , initDbConfig = Zlich
-  }
+snapshotConfig = fromFilePathConfig . toFilePath . unSnapshot
+
+-------------------------------------------------------------------------------
+-- cacheAction
+-------------------------------------------------------------------------------
+{-|
+Check to see if a cached data directory exists.
+
+If the file path does not exist the @initial@ config is used to start a @postgres@
+instance. After which the @action@ is applied, the data directory is cached
+and @postgres@ is shutdown.
+
+'cacheAction' 'mappend's a config to copy the cached data directory
+on startup onto the @initial@ config and returns it. In other words:
+
+@
+initialConfig <> configFromCachePath
+@
+
+@since 1.29.0.0
+-}
+cacheAction
+  :: FilePath
+  -- ^ Location of the data directory cache.
+  -> (DB -> IO ())
+  -- ^ @action@ to cache.
+  -> Config
+  -- ^ @initial@ 'Config'.
+  -> IO (Either StartError Config)
+cacheAction cachePath action config = do
+  fixCachePath <- fixPath cachePath
+  let result = config <> fromFilePathConfig fixCachePath
+  nonEmpty <- doesFileExist $ fixCachePath <> "/PG_VERSION"
+
+  case nonEmpty of
+    True -> pure $ pure result
+    False -> fmap join $ withConfig config $ \db -> do
+      action db
+      -- TODO see if parallel is better
+      throwIfNotSuccess id =<< stopPostgresGracefully db
+      createDirectoryIfMissing True fixCachePath
+
+      let snapshotCopyCmd = cpFlags <>
+            toDataDirectory db <> "/* " <> fixCachePath
+      system snapshotCopyCmd >>= \case
+        ExitSuccess -> pure $ pure result
+        x -> pure $ Left $ SnapshotCopyFailed snapshotCopyCmd x
