@@ -10,6 +10,7 @@ import           Data.List
 import qualified Data.Set as Set
 import           Data.String
 import qualified Database.PostgreSQL.Simple as PG
+import qualified Database.PostgreSQL.Simple.SqlQQ as PG
 import qualified Database.PostgreSQL.Simple.Options as Client
 import           Database.Postgres.Temp.Internal
 import           Database.Postgres.Temp.Internal.Config
@@ -105,6 +106,12 @@ data ConfigAndAssertion = ConfigAndAssertion
 memptyConfigAndAssertion :: ConfigAndAssertion
 memptyConfigAndAssertion = memptyConfigAndAssertion' "postgres"
 
+countDbs :: PG.Connection -> IO Int
+countDbs conn = PG.fromOnly . head <$> PG.query_ conn [PG.sql|
+  SELECT COUNT(*)
+  FROM pg_catalog.pg_database
+  |]
+
 memptyConfigAndAssertion' :: String -> ConfigAndAssertion
 memptyConfigAndAssertion' expectedDbName =
   let
@@ -375,8 +382,9 @@ happyPaths = describe "succeeds with" $ do
       either throwIO pure =<< withConfig (cacheConfig cacheInfo) assertConnection
 
   it "postgresql.conf append last wins" $
-    withConfig' (defaultPostgresConf [("fsync", "on")]) $ \db -> do
+    withConfig' (verboseConfig <> defaultPostgresConf [("fsync", "on")]) $ \db -> do
       toPostgresqlConfigFile db `shouldContain` "fsync=on"
+      withConn db $ \conn -> countDbs conn `shouldReturn` 3
 --
 -- Error Plans. Can't be combined. Just list them out inline since they can't be combined
 --
@@ -389,8 +397,8 @@ errorPaths = describe "fails when" $ do
     let invalidConfig = mempty
           { connectionTimeout = pure 0
           , connectionOptions = mempty
-                  { Client.dbname = pure "doesnotexist"
-                  }
+            { Client.dbname = pure "doesnotexist"
+            }
           }
 
     withConfig (defaultConfig <> invalidConfig) (const $ pure ())
@@ -400,11 +408,11 @@ errorPaths = describe "fails when" $ do
     let invalidConfig = mempty
           { connectionTimeout = pure maxBound
           , connectionOptions = mempty
-                  { Client.dbname = pure "doesnotexist"
-                  }
+            { Client.dbname = pure "doesnotexist"
+            }
           }
 
-    timeout 100000 (withConfig (defaultConfig <> invalidConfig) (const $ pure ()))
+    timeout 2000000 (withConfig (defaultConfig <> invalidConfig) (const $ threadDelay 10000000))
       `shouldReturn` Nothing
 {-
   it "throws StartPostgresFailed if the port is taken" $
@@ -471,10 +479,8 @@ errorPaths = describe "fails when" $ do
 
       withConfig nonEmptyFolderPlan mempty >>= \case
         Right () -> fail "Should not succeed"
-        Left ((InitDbFailed theOut theErr code)) -> do
+        Left ((InitDbFailed _ _ code)) ->
           code `shouldBe` ExitFailure 1
-          length theOut `shouldSatisfy` (> 0)
-          length theErr `shouldSatisfy` (> 0)
         Left err -> fail $ "Wrong type of error " <> show err
 
   it "invalid initdb options cause an error" $ do
@@ -544,12 +550,11 @@ withConfigSpecs = describe "withConfig" $ do
 withSnapshotSpecs :: Spec
 withSnapshotSpecs = describe "withSnapshot" $ do
   it "works" $ withConfig' defaultConfig $ \db -> do
-    -- TODO create a table
     withConn db $ \conn -> do
       _ <- PG.execute_ conn "BEGIN; CREATE TABLE foo ( id int );"
       void $ PG.execute_ conn "INSERT INTO foo (id) VALUES (1); END;"
 
-    either throwIO pure <=< withSnapshot Temporary db $ \snapshotDir -> do
+    either throwIO pure <=< withSnapshot db $ \snapshotDir -> do
       let theSnapshotConfig = defaultConfig <> snapshotConfig snapshotDir
           snapshotConfigAndAssert = ConfigAndAssertion theSnapshotConfig $ flip withConn $ \conn -> do
             oneAgain <- fmap (PG.fromOnly . head) $ PG.query_ conn "SELECT id FROM foo"
@@ -559,12 +564,79 @@ withSnapshotSpecs = describe "withSnapshot" $ do
         snapshotConfigAndAssert
         testSuccessfulConfig
 
+cacheActionSpecs :: Spec
+cacheActionSpecs = describe "cacheAction" $ do
+  it "creates the cache if it does not exist" $ do
+    let action db = withConn db $ \conn -> do
+          _ <- PG.execute_ conn "BEGIN; CREATE TABLE foo ( id int );"
+          void $ PG.execute_ conn "INSERT INTO foo (id) VALUES (1); END;"
+    withTempDirectory "/tmp" "tmp-postgres-cache-action" $ \cachePath -> do
+      let theFinalCachePath = cachePath <> "/cached"
+      cacheAction theFinalCachePath action defaultConfig >>= \case
+        Left err -> fail $ "cacheAction failed with:" <> show err
+        Right newCache -> do
+          nonEmpty <- doesFileExist $ theFinalCachePath <> "/PG_VERSION"
+          nonEmpty `shouldBe` True
+
+          -- Write a file and make sure it shows up in the data directory
+          writeFile (theFinalCachePath <> "/newFile.txt") "yes"
+
+          let
+            asserts db = do
+              doesFileExist (toDataDirectory db <> "/newFile.txt") `shouldReturn` True
+              withConn db $ \conn -> do
+                oneAgain <- fmap (PG.fromOnly . head) $ PG.query_ conn "SELECT id FROM foo"
+                oneAgain `shouldBe` (1 :: Int)
+            snapshotConfigAndAssert = ConfigAndAssertion newCache asserts
+
+          testWithTemporaryDirectory
+            snapshotConfigAndAssert
+            testSuccessfulConfig
+
+  it "doesn't create the cache if it exists" $ do
+    let action db = withConn db $ \conn -> do
+          _ <- PG.execute_ conn "BEGIN; CREATE TABLE foo ( id int );"
+          void $ PG.execute_ conn "INSERT INTO foo (id) VALUES (1); END;"
+    withTempDirectory "/tmp" "tmp-postgres-cache-action" $ \cachePath -> do
+      let theFinalCachePath = cachePath <> "/cached"
+      cacheAction theFinalCachePath action defaultConfig >>= \case
+        Left err -> fail $ "cacheAction failed with:" <> show err
+        Right newCache -> do
+          let nextAction db = withConn db $ \conn ->
+                void $ PG.execute_ conn "INSERT INTO foo (id) VALUES (2); END;"
+
+          cacheAction theFinalCachePath nextAction newCache >>= \case
+            Left err -> fail $ "cacheAction failed with:" <> show err
+            Right newCache1 -> do
+              let
+                asserts db = do
+                  withConn db $ \conn -> do
+                    oneAgain <- fmap (PG.fromOnly . head) $ PG.query_ conn "SELECT COUNT(*) FROM foo"
+                    oneAgain `shouldBe` (1 :: Int)
+                snapshotConfigAndAssert = ConfigAndAssertion newCache1 asserts
+
+              testWithTemporaryDirectory
+                snapshotConfigAndAssert
+                testSuccessfulConfig
+
+  it "fails if the cache director and data directory are the same" $ do
+    let action db = withConn db $ \conn -> do
+          _ <- PG.execute_ conn "BEGIN; CREATE TABLE foo ( id int );"
+          void $ PG.execute_ conn "INSERT INTO foo (id) VALUES (1); END;"
+    withTempDirectory "/tmp" "tmp-postgres-cache-action" $ \cachePath -> do
+      let theFinalCachePath = cachePath <> "/cached"
+      cacheAction theFinalCachePath action (defaultConfig { dataDirectory = Permanent theFinalCachePath } ) >>= \case
+        Left (SnapshotCopyFailed {}) -> pure ()
+        _ -> fail $ "cacheAction should have failed with SnapshotCopyFailed"
+
 spec :: Spec
 spec = do
   withConfigSpecs
 
   withSnapshotSpecs
 
+  cacheActionSpecs
+
   it "stopPostgres cannot be connected to" $ withConfig' defaultConfig $ \db -> do
     stopPostgres db `shouldReturn` ExitSuccess
     PG.connectPostgreSQL (toConnectionString db) `shouldThrow`
